﻿using GuruGuruSmf;
using System;
using System.Threading;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Reflection;
using System.Text;
using System.Windows.Forms;
using System.Xml.Serialization;

namespace TestPlayer
{
    public partial class MainForm : Form
    {
        private readonly IGuruGuruSmf4Api ggs = Ggs4Dll.GetInstance();
		private readonly Stopwatch stopwatch = new Stopwatch();
		private TextBox console;
		private SmfListBox ListBoxFiles;

		private readonly TabControl TabsTop;
		private readonly SplitContainer Split1, Split2;

		// デバイス設定関連
		private Label LabelDevice;
		private RadioButton RadioDevice1, RadioDevice2;
		private ComboBox ComboBoxDevices;
		private string nowDevice;

		// フォント設定関連
		private Label LabelFont;

		// 演奏オプション関連
		private CheckBox CheckLoop, CheckSkipBeginning, CheckSkipEnding,
			CheckMasterPitch, CheckNormalize, CheckExclusive;
		private NumericUpDown UpDownKey;
		private TrackBar TrackFadeIn, TrackFadeOut;

		// 演奏操作関連
		private TrackBar TrackTick, TrackVolume, TrackTempo, TrackMasterPitch;
		private Button ButtonPause;
		private bool doSeeking;
		private int seekValue;

		// 演奏中情報関連
		private Label LabelPlayTime, LabelTempo, LabelLoops, LabelTick;
		private int totalTick;
		private readonly System.Windows.Forms.Timer Timer;

		// 設定ファイルのパス
		private string PathSetting
        {
			get
            {
				var path = Assembly.GetExecutingAssembly().Location;
				var dir = Path.GetDirectoryName(path);
				return $"{dir}\\setting.xml";
            }
        }

        public MainForm()
        {
            InitializeComponent();

			Text = Util.Title;
			AllowDrop = true;
			KeyPreview = true;

            Load += Event_FormLoad;
            FormClosed += Event_FormClosed;
            DragEnter += Event_MainFormDragEnter;
			DragDrop += Event_MainFormDragDrop;
			KeyDown += Event_GlobalKeyDown;

            var timer = new System.Windows.Forms.Timer
            {
                Interval = 20,
            };
			timer.Tick += Event_TimerTick;
			Timer = timer;
			

			// スプリットパネルで画面分割
			var split1 = new SplitContainer
			{
				Dock = DockStyle.Fill,
				SplitterWidth = 3,
				Orientation = Orientation.Horizontal,
				FixedPanel = FixedPanel.Panel2,
				Panel1MinSize = 60,
			};
			Controls.Add(split1);
			Split1 = split1;

			// タブコントロールでプレイヤーと設定部を分ける
			var tabc = new TabControl
			{
				Dock = DockStyle.Fill,
			};
			tabc.TabPages.Add("プレイヤー");
			tabc.TabPages.Add("設定");
			split1.Panel1.Controls.Add(tabc);
			TabsTop = tabc;

			// プレイヤータブページを分割
			var split2 = new SplitContainer
			{
				Dock = DockStyle.Fill,
				SplitterWidth = 3,
				Orientation = Orientation.Vertical,
				FixedPanel = FixedPanel.Panel1,
				Panel2MinSize = 60,
			};
			tabc.TabPages[0].Controls.Add(split2);
			Split2 = split2;

			// 設定部
			var flow = new FlowLayoutPanel
			{
				Dock = DockStyle.Fill,
				FlowDirection = FlowDirection.TopDown,
				WrapContents = false,
				Padding = new Padding(4),
			};
			tabc.TabPages[1].Controls.Add(flow);

			var group1 = new GroupBox
			{
				Dock = DockStyle.Fill,
				Text = "デバイス",
				AutoSize = true,
			};
			flow.Controls.Add(group1);
			CreateGroupDevice(group1);

			var group2 = new GroupBox
			{
				Dock = DockStyle.Fill,
				Text = "フォント",
				AutoSize = true,
			};
			flow.Controls.Add(group2);
			CreateGroupFont(group2);

			CreatePanelConsole(split1.Panel2);
			CreatePanelFileList(split2.Panel1);
			CreatePanelController(split2.Panel2);
		}

		// コンソール用のパネル内部を作成
		private void CreatePanelConsole(Control parent)
        {
			var txt = new TextBox
			{
				Dock = DockStyle.Fill,
				Multiline = true,
				ScrollBars = ScrollBars.Vertical,
				ReadOnly = true,
				BackColor = SystemColors.HighlightText,
			};
			parent.Controls.Add(txt);
			console = txt;
		}
		// プレイヤーの左パネル内部を作成
		private void CreatePanelFileList(Control parent)
		{
			var list = new SmfListBox
			{
				Dock = DockStyle.Fill,
			};
			list.KeyDown += Event_SMFListKeyDown;
			list.MouseDoubleClick += Event_SMFListDoubleClick;

			parent.Controls.Add(list);
			ListBoxFiles = list;

		}
		// プレイヤー操作部を作成
		private void CreatePanelController(Control parent)
		{
			var tabs = new TabControl
			{
				Dock = DockStyle.Fill,
			};
			parent.Controls.Add(tabs);
			tabs.TabPages.Add("操作");
			tabs.TabPages.Add("オプション");

			var flow1 = new FlowLayoutPanel
			{
				Dock = DockStyle.Top,
				FlowDirection = FlowDirection.TopDown,
				WrapContents = false,
				AutoSize = true,
				Padding = new Padding(0, 0, 0, 8),
			};
			parent.Controls.Add(flow1);

			var flow2 = new FlowLayoutPanel
			{
				FlowDirection = FlowDirection.LeftToRight,
				AutoSize = true,
			};
			flow1.Controls.Add(flow2);

			var btn1 = new Button
			{
				Text = "Play",
				Width = 50,
				AutoSize = true,
			};
			btn1.Click += Event_PlayMusic;
			flow2.Controls.Add(btn1);
			var btn2 = new Button
			{
				Text = "Stop",
				Width = 50,
				AutoSize = true,
			};
			btn2.Click += Event_StopMusic;
			flow2.Controls.Add(btn2);
			var btn3 = new Button
			{
				Text = "Pause",
				Width = 50,
				AutoSize = true,
				Enabled = false,
			};
			flow2.Controls.Add(btn3);
			ButtonPause = btn3;
			btn3.Click += Event_PauseMusic;

			var flow3 = new FlowLayoutPanel
			{
				FlowDirection = FlowDirection.LeftToRight,
				WrapContents = false,
				AutoSize = true,
			};
			flow1.Controls.Add(flow3);

			int size = (int)Font.Size;
			int height = Font.Height;
			var label1 = new Label
			{
				Text = "---:--",
				Width = size * 5,
				Height = height,
			};
			flow3.Controls.Add(label1);
			LabelPlayTime = label1;
			var label2 = new Label
			{
				Text = "Tempo:---",
				Width = size * 8,
				Height = height,
			};
			flow3.Controls.Add(label2);
			LabelTempo = label2;
			var label3 = new Label
			{
				Text = "Loop:---",
				Width = size * 8,
				Height = height,
			};
			flow3.Controls.Add(label3);
			LabelLoops = label3;

			CreatePanelControllerValues(tabs.TabPages[0]);
			CreatePanelControllerOptions(tabs.TabPages[1]);
		}
		// プレイヤーの演奏操作部を作成
		private void CreatePanelControllerValues(Control parent)
        {
			var table = new TableLayoutPanel
			{
				Dock = DockStyle.Top,
				ColumnCount = 2,
				RowCount = 4,
				Padding = new Padding(2, 5, 2, 2),
				AutoSize = true,
			};
			parent.Controls.Add(table);

			void createTrack(out Label label, out TrackBar track, int row, string text, int value)
            {
				label = new Label
				{
					Anchor = AnchorStyles.Right,
					Text = text,
					AutoSize = true,
				};
				track = new TrackBar
				{
					Minimum = 0,
					Maximum = 127,
					TickFrequency = 8,
					Width = 200,
					Value = value,
					Tag = value,
				};
				table.Controls.Add(label, 0, row);
				table.Controls.Add(track, 1, row);
			}
			createTrack(out LabelTick, out TrackTick, 0, "--------", 0);
			createTrack(out _, out TrackVolume, 1, "Volume", 127);
			createTrack(out _, out TrackTempo, 2, "Tempo", 64);
			createTrack(out _, out TrackMasterPitch, 3, "Pitch", 64);

			TrackMasterPitch.Enabled = false;
			TrackTick.MouseDown += Event_TrackTickMouseDown;
			TrackTick.MouseUp += Event_TrackTickMouseUp;
			TrackVolume.ValueChanged += Event_TrackVolumeValueChanged;
			TrackTempo.ValueChanged += Event_TrackTempoValueChanged;
			TrackMasterPitch.ValueChanged += Event_TrackMasterPitchValueChanged;

			TrackVolume.MouseDown += Event_TrackMouseDown;
			TrackTempo.MouseDown += Event_TrackMouseDown;
			TrackMasterPitch.MouseDown += Event_TrackMouseDown;
		}
		// プレイヤーのオプション設定部を作成
		private void CreatePanelControllerOptions(Control parent)
        {
			var flow1 = new FlowLayoutPanel
			{
				Dock = DockStyle.Fill,
				FlowDirection = FlowDirection.TopDown,
				Padding = new Padding(2, 5, 2, 2),
				AutoSize = true,
			};
			parent.Controls.Add(flow1);

			var table1 = new TableLayoutPanel
			{
				ColumnCount = 2,
				RowCount = 4,
				AutoSize = true,
			};
			flow1.Controls.Add(table1);

			CheckBox createCheckBox(string text, int row, int col)
            {
				var check = new CheckBox
				{
					Text = text,
					AutoSize = true,
				};
				table1.Controls.Add(check, col, row);
				return check;
            }

			CheckLoop = createCheckBox("ループ再生", 0, 0);
			CheckSkipBeginning = createCheckBox("冒頭の空白をスキップ", 1, 0);
			CheckSkipEnding = createCheckBox("末尾の空白をスキップ", 2, 0);
			CheckNormalize = createCheckBox("ノーマライズ", 0, 1);
			CheckExclusive = createCheckBox("エクスクルーシブを送信", 1, 1);
			CheckMasterPitch = createCheckBox("マスターピッチ", 2, 1);

			var flow2 = new FlowLayoutPanel
			{
				FlowDirection = FlowDirection.LeftToRight,
				AutoSize = true,
			};
			table1.Controls.Add(flow2, 0, 3);
			table1.SetColumnSpan(flow2, 1);

			var label1 = new Label
			{
				Text = "キー",
				Anchor = AnchorStyles.Left,
				AutoSize = true,
			};
			flow2.Controls.Add(label1);
			var updown = new NumericUpDown
			{
				Width = (int)(Font.Size * 4) + 5,
				Minimum = -24,
				Maximum = 24,
				Value = 0,
			};
			flow2.Controls.Add(updown);
			UpDownKey = updown;

			var table2 = new TableLayoutPanel
			{
				ColumnCount = 2,
				RowCount = 2,
				AutoSize = true,
			};
			flow1.Controls.Add(table2);

			void createTrack(out Label label, out TrackBar track, int row, string text, int value)
			{
				label = new Label
				{
					Anchor = AnchorStyles.Right,
					Text = text,
					AutoSize = true,
				};
				track = new TrackBar
				{
					Minimum = 0,
					Maximum = 127,
					TickFrequency = 8,
					Width = 200,
					Value = value,
					Tag = value,
				};
				table2.Controls.Add(label, 0, row);
				table2.Controls.Add(track, 1, row);
			}
			createTrack(out _, out TrackFadeIn, 0, "Fade In", 0);
			createTrack(out _, out TrackFadeOut, 1, "Fade Out", 0);

			TrackFadeIn.MouseDown += Event_TrackMouseDown;
			TrackFadeOut.MouseDown += Event_TrackMouseDown;
		}
		// デバイス設定部分を作成
		private void CreateGroupDevice(Control parent)
        {
			var panel = new TableLayoutPanel
			{
				Dock = DockStyle.Fill,
				AutoSize = true,
				ColumnCount = 2,
				RowCount = 4,
				Padding = new Padding(2),
			};
			parent.Controls.Add(panel);

			var label = new Label
			{
				Text = "現在のデバイス: DirectMusic",
				AutoSize = true,
				Anchor = AnchorStyles.Left,
				Margin = new Padding(2, 2, 0, 6),
			};
			panel.Controls.Add(label, 0, 0);
			panel.SetColumnSpan(label, 2);
			LabelDevice = label;

			var radio1 = new RadioButton
			{
				Text = "DirectMusic",
				AutoSize = true,
				Checked = true,
				Anchor = AnchorStyles.Left,
			};
			panel.Controls.Add(radio1, 0, 1);
			RadioDevice1 = radio1;
			var radio2 = new RadioButton
			{
				Text = "Windows API",
				AutoSize = true,
				Anchor = AnchorStyles.Left,
			};
			panel.Controls.Add(radio2, 0, 2);
			RadioDevice2 = radio2;

			var combo = new ComboBox
			{
				DropDownStyle = ComboBoxStyle.DropDownList,
				AutoSize = true,
				Anchor = AnchorStyles.Left,
			};
			panel.Controls.Add(combo, 1, 2);
			ComboBoxDevices = combo;

			var button = new Button
			{
				Text = "設定！",
				AutoSize = true,
			};
			panel.Controls.Add(button, 0, 3);
			button.Click += Event_SetDevice;
		}
		// フォント設定部分を作成
		private void CreateGroupFont(Control parent)
        {
			var flow = new FlowLayoutPanel
			{
				Dock = DockStyle.Fill,
				AutoSize = true,
				FlowDirection = FlowDirection.TopDown,
				WrapContents = false,
				Padding = new Padding(2),
			};
			parent.Controls.Add(flow);
			var label = new Label
			{
				Text = "現在のフォント",
				AutoSize = true,
				Margin = new Padding(4, 2, 0, 6),
			};
			flow.Controls.Add(label);
			LabelFont = label;
			var button = new Button
			{
				Text = "設定ダイアログ",
				AutoSize = true,
			};
			flow.Controls.Add(button);
			button.Click += Event_FontDialog;

		}


		// 演奏！
		private void PlayMusic()
        {
			try
			{
				console.Clear();
				GgsError err;

				var smf = (FileDataSmf)ListBoxFiles.SelectedItem;
				if (smf == null) return;

				ggs.Stop(0);

				// 作者の用途的にファイルが置き換わる可能性があるので毎回読み直します
				stopwatch.Reset();
				stopwatch.Start();
				err = ggs.AddListFromFile(smf.FilePath, 0, 0);
				stopwatch.Stop();
				if (err != GgsError.NoError)
				{
					console.AppendText($"SMF読み込み失敗 : {smf.Name} : {err} : {Util.TickToMilliseconds(stopwatch.ElapsedTicks)}\r\n");
					return;
				}

				var sb = new StringBuilder();
				sb.AppendLine($"SMF読み込み成功 : {smf.Name} : {Util.TickToMilliseconds(stopwatch.ElapsedTicks)}");
				sb.AppendLine();
				WriteSmfInfo(sb);
				sb.AppendLine();

				// 演奏！
				GetPlayOption(out int option, out int fadein, out int key);
				err = ggs.Play(option, 0, 0, fadein, key);
				if (err == GgsError.NoError)
				{
					sb.AppendLine($"演奏を開始します");
					Timer.Start();
					TrackMasterPitch.Enabled = CheckMasterPitch.Checked;
					ButtonPause.Text = "Pause";
					ButtonPause.Enabled = true;
				}
				else
				{
					sb.AppendLine($"演奏に失敗しました: {err}");
				}
				console.AppendText(sb.ToString());
			}
			catch (Exception ex)
			{
				console.Text = $"Playで例外が発生しました: {ex.Message}\r\n{ex.StackTrace}";
				Timer.Stop();
			}
		}
		// 停止！
		private void StopMusic()
        {
			var fade = 5000 * TrackFadeOut.Value / 127;
			ggs.Stop(fade);
			ButtonPause.Text = "Pause";
			ButtonPause.Enabled = false;
		}

		private void GetPlayOption(out int option, out int fadein, out int key)
        {
			option = 0;
			if (CheckLoop.Checked) option |= PlayOption.Loop;
			if (CheckSkipBeginning.Checked) option |= PlayOption.SkipBeginning;
			if (CheckSkipEnding.Checked) option |= PlayOption.SkipEnding;
			if (CheckMasterPitch.Checked) option |= PlayOption.MasterPitch;
			if (CheckNormalize.Checked) option |= PlayOption.Normalize;
			if (CheckExclusive.Checked) option |= PlayOption.Exclusive;
			fadein = 5000 * TrackFadeIn.Value / 127;
			key = (int)UpDownKey.Value;
        }
		// SMFの情報を書き込み
		private void WriteSmfInfo(StringBuilder sb)
        {
			ggs.GetSmfInformation(out SmfInformation smfInfo, 0);
			totalTick = smfInfo.TotalTick;

			string toStr(int time, int tick)
			{
				int min = time / 60000;
				int sec = (time / 1000) % 60;
				int mil = time % 1000;
				return $"{min,2}m{sec:D2}s{mil:D3} - {tick}";
			}

			sb.AppendLine($"Total    : {toStr(smfInfo.TotalTime, smfInfo.TotalTick)}");
			sb.AppendLine($"FirstNote: {toStr(smfInfo.FirstNoteTime, smfInfo.FirstNoteTick)}");
			sb.AppendLine($"LastNote : {toStr(smfInfo.LastNoteTime, smfInfo.LastNoteTick)}");
			string s;
			if (smfInfo.LoopTick == -1)
			{
				s = "なし";
			}
			else
			{
				s = toStr(smfInfo.LoopTime, smfInfo.LoopTick);
			}
			sb.AppendLine($"CC#111   : {s}");
		}

		private void SetLabelFontText()
        {
			LabelFont.Text = $"現在のフォント: {Font.Name}　{Font.SizeInPoints}pt";
        }

		private void OpenDevice()
        {
			var r1 = RadioDevice1.Checked;
			var r2 = RadioDevice2.Checked;
			var item = (WinMMDeviceData)ComboBoxDevices.SelectedItem;
			var windev = -1;
			var name = "";
			if (item != null)
			{
				windev = item.Id;
				name = item.Name;
			}
			string mes;
			string devname;
			GgsError err;

			stopwatch.Reset();
			stopwatch.Start();
			if (r1)
			{
				err = ggs.OpenDevice(Device.DirectMusic, Handle);
				devname = "DirectMusic";
			}
			else if (r2)
			{
				err = ggs.OpenDevice(windev, Handle);
				devname = name;
			}
			else
            {
				stopwatch.Stop();
				throw new Exception("ラジオボタンがチェックされてません。");
            }
			stopwatch.Stop();

			var t = Util.TickToMilliseconds(stopwatch.ElapsedTicks);
			if (err == GgsError.NoError)
			{
				mes = $"デバイスを開きました : {devname} : {t}\r\n";
				nowDevice = devname;
			}
			else
			{
				mes = $"デバイスを開くのに失敗しました : {devname} : {err} : {t}\r\n";
				nowDevice = null;
				devname = "(なし)";
			}
			LabelDevice.Text = $"現在のデバイス: {devname}";
			console.AppendText(mes);
		}

		private bool GetIsPlaying()
        {
			var stat = ggs.GetPlayerStatus();
			return stat.State == PlayerState.Play || stat.State == PlayerState.Pause;
		}

		// イベントハンドラー
		private void Event_FormLoad(object sender, EventArgs e)
        {
			// 設定ファイルの読み込み
			Setting setting;
			try
            {
				setting = Setting.Load(PathSetting);
				//setting = null;
            }
			catch(Exception)
            {
				setting = null;	// しっぱい
            }

			// フォームの位置などを設定
			if (setting != null)
            {
				var form = setting.Form;
				Left = form.Left;
				Top = form.Top;
				Width = form.Width;
				Height = form.Height;
				Split1.SplitterDistance = form.Split1;
				Split2.SplitterDistance = form.Split2;

				Font = new Font(setting.Font.Name, setting.Font.Size);
            }
			else
            {
				Height = Font.Height * 40;
				Width = (int)(Font.Size * 55);
				Split1.SplitterDistance = Height - (Font.Height * 12);
				Split2.SplitterDistance = (int)(Font.Size * 16);
			}

			// デバイスのリストを取得してコンボボックスに入れる
			var devs = new WinMMDeviceList();
			devs.ApplyTo(ComboBoxDevices);	// ここでコンボボックスのサイズを設定してる

			// デバイスの設定
			if (setting != null && setting.Device != null && setting.Device != "DirectMusic")
			{
				// DirectMusicがデフォルトなので、WindowsAPIの中から探す
				RadioDevice2.Checked = true;

				foreach (WinMMDeviceData dev in ComboBoxDevices.Items)
                {
					if (dev.Name == setting.Device)
                    {
						ComboBoxDevices.SelectedItem = dev;
                    }
                }
			}

			OpenDevice();

			SetLabelFontText();

			// コマンドラインのチェック
			string[] args = Environment.GetCommandLineArgs();
			if (args.Length > 1)
            {
				if (Path.GetExtension(args[1]).ToLower() == ".mid")
				{
					var smf = new FileDataSmf(args[1]);
					ListBoxFiles.Add(smf);
					ListBoxFiles.SelectedItem = smf;
					PlayMusic();
				}
			}

		}
		private void Event_FormClosed(object sender, EventArgs e)
        {
			ggs.CloseDevice();

			// 設定保存
			var setting = new Setting();
			setting.Initialize();
			setting.Form.SetRect(Left, Top, Width, Height);
			setting.Form.SetSplit(Split1.SplitterDistance, Split2.SplitterDistance);
			setting.Font.Set(Font.Name, Font.Size);
			setting.Device = nowDevice;

            try
            {
				setting.Save(PathSetting);
			}
			catch (Exception)
            {
				// ここで例外を飛ばされても困るので知らんぷり
			}
        }
		private void Event_PlayMusic(object sender, EventArgs e)
        {
			PlayMusic();
		}
		private void Event_StopMusic(object sender, EventArgs e)
		{
			StopMusic();
		}
		private void Event_PauseMusic(object sender, EventArgs e)
		{
			var stat = ggs.GetPlayerStatus();
			switch (stat.State)
            {
				case PlayerState.Play:
					ggs.Pause();
					ButtonPause.Text = "Restart";
					break;
				case PlayerState.Pause:
					ggs.Restart();
					ButtonPause.Text = "Pause";
					break;
				default:
					break;
            }
		}
		private void Event_TimerTick(object sender, EventArgs e)
        {
			var stat = ggs.GetPlayerStatus();

			var min = stat.Time / 60000;
			var sec = (stat.Time / 1000) % 60;
			LabelPlayTime.Text = $"{min}:{sec:D2}";
			LabelTempo.Text = $"Tempo:{stat.Tempo}";
			LabelLoops.Text = $"Loop:{stat.LoopCount}";
			LabelTick.Text = $"{stat.Tick,8}";
			if (!doSeeking && totalTick != 0)
			{
				TrackTick.Value = (127 * stat.Tick / totalTick);
			}
			if (stat.State == PlayerState.Stop || stat.State == PlayerState.Suspended)
			{
				Timer.Stop();
				ButtonPause.Enabled = false;
			}
		}

		// 演奏操作
		private void Event_TrackTickMouseDown(object sender, MouseEventArgs e)
		{
			doSeeking = true;
			seekValue = TrackTick.Value;
		}
		private void Event_TrackTickMouseUp(object sender, MouseEventArgs e)
		{
			if (GetIsPlaying() && seekValue != TrackTick.Value)
            {
				int seek = TrackTick.Value * totalTick / 127;
				ButtonPause.Text = "Pause";
				GetPlayOption(out int option, out int fadein, out int key);
				ggs.Play(option, 0, seek, fadein, key);
			}
			doSeeking = false;
		}

		private void Event_TrackVolumeValueChanged(object sender, EventArgs e)
		{
			int v = TrackVolume.Value - 127;
			ggs.SetMasterVolume(v);
		}
		private void Event_TrackTempoValueChanged(object sender, EventArgs e)
        {
			int v = TrackTempo.Value - 64;	// -64～+63
			double x;
			if (v > 0)
            {
				x = 1 + (double)v / 63 * 2;	// 最大3
            }
			else
            {
				x = 1 + (double)v / 64 * 0.7;	// 最小0.3
            }
			ggs.SetMasterTempo((float)x);
        }
		private void Event_TrackMasterPitchValueChanged(object sender, EventArgs e)
		{
			int v = (TrackMasterPitch.Value - 64) * 8192 / 64;
			ggs.SetMasterPitch(v);
		}

		private void Event_TrackMouseDown(object sender, MouseEventArgs e)
        {
			// 右ボタンが押されたらそれぞれの初期位置へ
			var track = (TrackBar)sender;
			if (e.Button == MouseButtons.Right)
            {
				track.Value = (int)track.Tag;   // とりあえずTagを利用
            }
        }

		// ファイルリスト操作
		private void Event_SMFListKeyDown(object sender, KeyEventArgs e)
        {
			switch (e.KeyData)
			{
				// 演奏
				case Keys.Enter:
					PlayMusic();
					e.SuppressKeyPress = true;
					break;
				// 停止
				case Keys.Escape:
					StopMusic();
					e.SuppressKeyPress = true;
					break;
				default:
					break;
			}
		}
		private void Event_SMFListDoubleClick(object sender, MouseEventArgs e)
        {
			var list = (SmfListBox)sender;
			var idx = list.IndexFromPoint(e.Location);
			if (idx == ListBox.NoMatches) return;
			list.SelectedIndex = idx;
			Event_PlayMusic(sender, e);
        }

		// 設定関連
		private void Event_SetDevice(object sender, EventArgs e)
		{
			console.Clear();
			OpenDevice();
		}
		private void Event_FontDialog(object sender, EventArgs e)
		{
			var dlg = new FontDialog
			{
				Font = Font,
				Color = ForeColor,
				MaxSize = 32,
				MinSize = 0,
				FontMustExist = true,
				AllowVerticalFonts = false,
				ShowColor = false,
				ShowEffects = false,
				FixedPitchOnly = false,
				AllowVectorFonts = true,
			};
			if (dlg.ShowDialog() == DialogResult.OK)
            {
				Font = dlg.Font;
				SetLabelFontText();
            }
		}

		// 大域ショートカットキー処理
		private void Event_GlobalKeyDown(object sender, KeyEventArgs e)
        {
			switch (e.KeyData)
			{
				// SMFリストにフォーカスを移す
				case Keys.Control | Keys.F:
					if (TabsTop.SelectedIndex != 0) TabsTop.SelectedIndex = 0;
					ListBoxFiles.Focus();
					e.Handled = true;
					break;
				default:
					break;
			}

		}

		// ドラッグ＆ドロップ
		private void Event_MainFormDragEnter(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(DataFormats.FileDrop))
            {
                e.Effect = DragDropEffects.Copy;
            }
        }
		private void Event_MainFormDragDrop(object sender, DragEventArgs e)
		{
			console.Clear();
			console.AppendText("ファイルがドロップされました\r\n");
			ListBoxFiles.BeginUpdate();
			FileDataSmf first = null;
			string[] drags = (string[])e.Data.GetData(DataFormats.FileDrop);
			var sb = new StringBuilder();
			int countSmf = 0;
			foreach (string fileName in drags)
			{
				string ext = Path.GetExtension(fileName);
				ext = ext.ToLower();
				if (ext == ".dls")
				{
					var err = ggs.AddDls(fileName);
					if (err == GgsError.NoError)
                    {
						sb.AppendLine($"{fileName} > 読み込み成功");
					}
					else
                    {
						sb.AppendLine($"{fileName} > 読み込み失敗 : {err}");
					}
				}
				else if (ext == ".mid")
				{
					var smf = new FileDataSmf(fileName);
					if (ListBoxFiles.Add(smf))
                    {
						sb.AppendLine($"{fileName} > 追加");
						++countSmf;
						if (first == null) first = smf;
					}
					else
                    {
						sb.AppendLine($"{fileName} > スキップ");
					}
				}
			}
			// 読み込んだ最初のSMFを選択
			if (first != null) ListBoxFiles.SelectedItem = first;
			sb.AppendLine($"{countSmf}のSMFが追加されました");
			console.AppendText(sb.ToString());
			ListBoxFiles.EndUpdate();
		}

		// Windowsメッセージの処理
        protected override void WndProc(ref Message m)
        {
			base.WndProc(ref m);
			switch (m.Msg)
            {
				// 送られて来たSMFを再生
				case WIN.WM_COPYDATA:
					var cds = (WIN.COPYDATASTRUCT)m.GetLParam(typeof(WIN.COPYDATASTRUCT));
					var smf = new FileDataSmf(cds.lpData);
					ListBoxFiles.Add(smf, out int index);
					ListBoxFiles.SelectedIndex = index;
					PlayMusic();
					break;
            }
		}
	}

	public class Setting
    {
		public class MainForm
        {
			public int Left { get; set; }
			public int Top { get; set; }
			public int Width { get; set; }
			public int Height { get; set; }
			public int Split1 { get; set; }
			public int Split2 { get; set; }

			public void SetRect(int left, int top, int width, int height)
            {
				Left = left;
				Top = top;
				Width = width;
				Height = height;
            }
			public void SetSplit(int split1, int split2)
            {
				Split1 = split1;
				Split2 = split2;
            }
        }
		public class FontData
        {
			public string Name { get; set; }
			public float Size { get; set; }
			public void Set(string name, float size)
            {
				Name = name;
				Size = size;
            }
        }


		public MainForm Form { get; set; }
		public FontData Font { get; set; }
		public string Device { get; set; }


		public void Initialize()
        {
			Form = new MainForm();
			Font = new FontData();
        }

		public void Save(string path)
        {
			var xml = new XmlSerializer(typeof(Setting));
			using (var sw = new StreamWriter(path, false, new UTF8Encoding(false)))
			{
				xml.Serialize(sw, this);
			}
		}
		public static Setting Load(string path)
        {
			var xml = new XmlSerializer(typeof(Setting));
			using (var sr = new StreamReader(path, new UTF8Encoding(false)))
			{
				return (Setting)xml.Deserialize(sr);
			}
        }
    }



}
