﻿using System.Collections.Generic;
using System.IO;
using System.Windows.Forms;

namespace TestPlayer
{
    public class SmfListBox : ListBox
    {
        private readonly SortedDictionary<string, FileDataSmf> _dicPath = new SortedDictionary<string, FileDataSmf>();
        

        public bool Add(FileDataSmf smf)
        {
            // 重複チェック
            if (_dicPath.ContainsKey(smf.FilePath))
            {
                return false;
            }

            Items.Add(smf);
            _dicPath.Add(smf.FilePath, smf);
            return true;
        }
        public bool Add(FileDataSmf smf, out int index)
        {
            // 重複チェック
            if (_dicPath.ContainsKey(smf.FilePath))
            {
                var smf1 = _dicPath[smf.FilePath];
                index = Items.IndexOf(smf1);
                return false;
            }

            index = Items.Add(smf);
            _dicPath.Add(smf.FilePath, smf);
            return true;
        }

        private void SortByFilePath()
        {
            BeginUpdate();
            var sel = SelectedItem;
            Items.Clear();
            foreach (var kv in _dicPath)
            {
                Items.Add(kv.Value);
            }
            SelectedItem = sel;
            EndUpdate();
        }

        private void SelectedItem_Delete()
        {
            var idx = SelectedIndex;
            var item = SelectedItem;
            if (item == null) return;
            Items.Remove(item);
            _dicPath.Remove(((FileDataSmf)item).FilePath);
            if (Items.Count > idx)
            {
                SelectedIndex = idx;
            }
            else
            {
                if (idx > 0)
                {
                    SelectedIndex = idx - 1;
                }
            }
        }
        private void SelectedItem_MoveUp()
        {
            var idx = SelectedIndex;
            if (idx < 1) return;
            var temp = Items[idx];
            Items[idx] = Items[idx - 1];
            Items[idx - 1] = temp;
            SelectedIndex = idx - 1;
            Refresh();
        }
        private void SelectedItem_MoveDown()
        {
            var idx = SelectedIndex;
            if (idx >= Items.Count - 1) return;
            var temp = Items[idx];
            Items[idx] = Items[idx + 1];
            Items[idx + 1] = temp;
            SelectedIndex = idx + 1;
            Refresh();
        }

        protected override void OnKeyDown(KeyEventArgs e)
        {
            base.OnKeyDown(e);
            switch (e.KeyData)
            {
                // 選択アイテムを削除
                case Keys.Delete:
                    SelectedItem_Delete();
                    e.SuppressKeyPress = true;
                    break;
                // 全アイテムを削除
                case Keys.Control | Keys.Shift | Keys.Delete:
                    Items.Clear();
                    _dicPath.Clear();
                    e.SuppressKeyPress = true;
                    break;
                // 選択アイテムを上に移動
                case Keys.Shift | Keys.Up:
                    SelectedItem_MoveUp();
                    e.SuppressKeyPress = true;
                    break;
                // 選択アイテムを下に移動
                case Keys.Shift | Keys.Down:
                    SelectedItem_MoveDown();
                    e.SuppressKeyPress = true;
                    break;
                // フルパスでソート
                case Keys.Control | Keys.Z:
                    SortByFilePath();
                    e.SuppressKeyPress = true;
                    break;
                default:
                    break;
            }
        }
    }

    public class FileDataSmf
    {
        public string FilePath { get; private set; }
        public string Name { get; private set; }

        public FileDataSmf(string path)
        {
            FilePath = path;
            Name = Path.GetFileName(path);
        }

        public override string ToString()
        {
            return Name;
        }
    }

}
