﻿using System;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;

namespace TestPlayer
{
    public class WIN
    {
        [DllImport("User32.dll")]
        public static extern IntPtr FindWindow(string lpClassName, string lpWindowName);

        [DllImport("User32.dll")]
        public static extern IntPtr SendMessageW(IntPtr hWnd, uint Msg, uint wParam, ref COPYDATASTRUCT lParam);

        public struct COPYDATASTRUCT
        {
            public IntPtr dwData;
            public uint cbData;
            public string lpData;
        }
        public const int WM_COPYDATA = 0x4A;
    }

    public class Lock : IDisposable
    {
        private Mutex _mutex;
        private bool isOwned;
        public bool TryLock(string name)
        {
            _mutex = new Mutex(false, name);
            try
            {
                isOwned = _mutex.WaitOne(0, false);
            }
            catch(AbandonedMutexException)
            {
                isOwned = true;
            }
            return isOwned;
        }

        public void Dispose()
        {
            if (isOwned) _mutex.ReleaseMutex();
            _mutex.Close();
        }
    }

    public class Util
    {
        public static string Title
        {
            get
            {
                var loc = Assembly.GetExecutingAssembly().Location;
                var ver = FileVersionInfo.GetVersionInfo(loc);
                return $"GuruGuruSMF TestPlayer {ver.FileVersion}";
            }
        }

        public static string TickToMilliseconds(long tick)
        {
            decimal t = (decimal)tick * 1000 / Stopwatch.Frequency;
            return $"{t:F4}ms";
        }

        // 別ウインドウにコマンドラインのファイル名を送信する
        public static void SendArgAnotherWindow()
        {
            string[] args = Environment.GetCommandLineArgs();
            if (args.Length < 2) return;
            var text = args[1];
            if (Path.GetExtension(text).ToLower() != ".mid") return;

            var hwnd = WIN.FindWindow(null, Title);
            if (hwnd == IntPtr.Zero) return;

            var cds = new WIN.COPYDATASTRUCT
            {
                dwData = IntPtr.Zero,
                cbData = (uint)Encoding.Unicode.GetByteCount(text),
                lpData = text,
            };
            WIN.SendMessageW(hwnd, WIN.WM_COPYDATA, 0, ref cds);
        }
    }


}
