﻿using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace TestPlayer
{
	public class WinMMDeviceData
	{
		public int Id { get; }
		public string Name { get; }

		public WinMMDeviceData(int id, string name)
		{
			Id = id;
			Name = name;
		}

		public override string ToString()
		{
			return Name;
		}
	}

	public class WinMMDeviceList : List<WinMMDeviceData>
	{
		[StructLayout(LayoutKind.Sequential)]
		struct MIDIOUTCAPS
		{
			public short wMid;
			public short wPid;
			public int vDriverVersion;
			[MarshalAs(UnmanagedType.ByValTStr, SizeConst = 32)]
			public string szPname;
			public short wTechnology;
			public short wVoices;
			public short wNotes;
			public short wChannelMask;
			public uint dwSupport;
		}

		[DllImport("winmm.dll", EntryPoint = "midiOutGetNumDevs")]
		private static extern uint MidiOutGetNumDevs();
		[DllImport("winmm.dll", EntryPoint = "midiOutGetDevCapsA")]
		private static extern uint MidiOutGetDevCapsA(
			int uDeviceID, ref MIDIOUTCAPS lpMidiOutCaps, uint cbMidiOutCaps);

		public WinMMDeviceList()
		{
			uint devs = MidiOutGetNumDevs();
			MIDIOUTCAPS caps = new MIDIOUTCAPS();
			for (int i = -1; i < devs; ++i)
			{
				MidiOutGetDevCapsA(i, ref caps, (uint)Marshal.SizeOf(caps));
				Add(new WinMMDeviceData(i, caps.szPname));
			}
		}

		public void ApplyTo(ComboBox combo)
		{
			int width1 = 0;
			combo.Items.Clear();
			foreach (var dev in this)
			{
				combo.Items.Add(dev);
				int width2 = TextRenderer.MeasureText(dev.Name, combo.Font).Width;
				if (width1 < width2) width1 = width2;
			}
			// デバイスがないとかなさそうですけど念のため
			if (Count > 0)
			{
				combo.SelectedIndex = 0;
				combo.Width = width1 + 20;  // ちょっと大きめに
			}
		}


	}

}
